@@ -1,33 +1,11 @@
-"""
-Copyright 2021 Demetry Pascal (forked from Ryan (Mohammad) Solgi)
 
-Permission is hereby granted, free of charge, to any person obtaining a copy of 
-this software and associated documentation files (the "Software"), to deal in 
-the Software without restriction, including without limitation the rights to use, 
-copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the 
-Software, and to permit persons to whom the Software is furnished to do so, 
-subject to the following conditions:
-
-The above copyright notice and this permission notice shall be included in all
-copies or substantial portions of the Software.
-
-THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
-IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
-FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
-THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
-LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
-OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
-SOFTWARE.
-"""
-
-
-from typing import Callable, List, Tuple, Optional, Dict, Any, Union, Sequence, Set
+from typing import Callable, List, Tuple, Optional, Dict, Any, Union, Sequence, Literal, Iterable
+from typing_extensions import TypeAlias
 
 import collections
 import warnings
 import operator
 
-
 import sys
 import time
 import random
@@ -35,55 +13,71 @@ import math
 
 import numpy as np
 
+from OppOpPopInit.initialiser import CreatorFunc
+from OppOpPopInit.oppositor import OppositorFunc
+
 #region INTERNAL IMPORTS
 
-from .aliases import array1D, array2D
+from .utils.aliases import array1D, array2D
 
-from .classes import AlgorithmParams, Generation, MiddleCallbackData, GAResult, GenerationConvertible
+from .data_types.aliases import FunctionToMinimize, SetFunctionToMinimize
+from .data_types.algorithm_params import AlgorithmParams
+from .data_types.generation import GenerationConvertible, Generation
+from .data_types.result import GAResult
 
-from .initializer import Population_initializer
-from .plotting_tools import plot_pop_scores, plot_several_lines
+from .population_initializer import get_population_initializer, PopulationModifier
+from .utils.plotting import plot_pop_scores, plot_several_lines
 
-from .utils import can_be_prob, is_numpy, is_current_gen_number, fast_min, random_indexes_pair
+from .utils.funcs import can_be_prob, is_numpy, is_current_gen_number, fast_min, random_indexes_pair
 
-from .callbacks import MiddleCallbackFunc, CallbackFunc
+from .callbacks.data import MiddleCallbackData
+from .callbacks import MiddleCallbackFunc, SimpleCallbackFunc
+
+#endregion
+
+#region ALIASES
+
+VARIABLE_TYPE: TypeAlias = Literal['int', 'real', 'bool']
+"""
+the variable type for a given or all dimension, determines the values discretion:
+    real: double numbers
+    int: integer number only
+    bool: in the fact is integer with bounds [0, 1]
+"""
 
 #endregion
 
 
-class geneticalgorithm2:
-    
+class GeneticAlgorithm2:
     """
-    Genetic Algorithm (Elitist version) for Python
-    
-    An implementation of elitist genetic algorithm for solving problems with
-    continuous, integers, or mixed variables.
-    
-    repo path https://github.com/PasaOpasen/geneticalgorithm2
+    Genetic algorithm optimization process
     """
     
     default_params = AlgorithmParams()
     PROGRESS_BAR_LEN = 20
+    """max count of symbols in the progress bar"""
 
     @property
     def output_dict(self):
         warnings.warn(
-            "'output_dict' is deprecated and will be removed at version 7 \n use 'result' instead"
+            "'output_dict' is deprecated and will be removed at version 7 \n"
+            "use 'result' instead"
         )
         return self.result
 
     @property
     def needs_mutation(self) -> bool:
+        """whether the mutation is required"""
         return self.prob_mut > 0 or self.prob_mut_discrete > 0
 
     #region INIT
 
     def __init__(
         self,
-        function: Callable[[array1D], float],
+        function: FunctionToMinimize,
 
         dimension: int,
-        variable_type: Union[str, Sequence[str]] = 'bool',
+        variable_type: Union[VARIABLE_TYPE, Sequence[VARIABLE_TYPE]] = 'bool',
         variable_boundaries: Optional[Union[array2D, Sequence[Tuple[float, float]]]] = None,
 
         variable_type_mixed=None,
@@ -92,43 +86,40 @@ class geneticalgorithm2:
         algorithm_parameters: Union[AlgorithmParams, Dict[str, Any]] = default_params
     ):
         """
+        initializes the GA object and performs main checks
+
         Args:
-            function <Callable[[np.ndarray], float]> - the given objective function to be minimized
-            #NOTE: This implementation minimizes the given objective function.
-            (For maximization multiply function by a negative sign: the absolute
-            value of the output would be the actual objective function)
+            function: the given objective function to be minimized
+            dimension: the number of decision variables, the population samples dimention
 
-            dimension <integer> - the number of decision variables
+            variable_type: string means the variable type for all variables,
+                for mixed types use sequence of strings of type for each variable
 
-            variable_type <string> - 'bool' if all variables are Boolean;
-            'int' if all variables are integer; and 'real' if all variables are
-            real value or continuous. For mixed types use sequence of string of type for each variable
-
-            variable_boundaries <Optional[Union[np.ndarray, Sequence[Tuple[float, float]]]]> - Default None; leave it
-            None if variable_type is 'bool'; otherwise provide an array of tuples
-            of length two as boundaries for each variable;
-            the length of the array must be equal dimension. For example,
-            np.array([0,100],[0,200]) determines lower boundary 0 and upper boundary 100 for first
-            and upper boundary 200 for second variable where dimension is 2.
+            variable_boundaries: leave it None if variable_type is 'bool';
+                otherwise provide a sequence of tuples of length two as boundaries for each variable;
+                the length of the array must be equal dimension.
+                For example, ([0,100], [0,200]) determines
+                    lower boundary 0 and upper boundary 100 for first
+                    and upper boundary 200 for second variable
+                    and dimension must be 2.
 
             variable_type_mixed -- deprecated
 
-            function_timeout <float> - if the given function does not provide
-            output before function_timeout (unit is seconds) the algorithm raise error.
-            For example, when there is an infinite loop in the given function. `None` means disabling
+            function_timeout: if the given function does not provide
+                output before function_timeout (unit is seconds) the algorithm raises error.
+                For example, when there is an infinite loop in the given function.
+                `None` means disabling
 
-            algorithm_parameters <Union[AlgorithmParams, Dict[str, Any]]>:
-                @ max_num_iteration <int> - stoping criteria of the genetic algorithm (GA)
-                @ population_size <int>
-                @ mutation_probability <float in [0,1]>
-                @ elit_ratio <float in [0,1]>
-                @ crossover_probability <float in [0,1]>
-                @ parents_portion <float in [0,1]>
-                @ crossover_type <string/function> - Default is 'uniform'; 'one_point' or 'two_point' (not only) are other options
-                @ mutation_type <string/function> - Default is 'uniform_by_x'; see GitHub to check other options
-                @ mutation_discrete_type <string/function> - mutation type for discrete variables
-                @ selection_type <string/function> - Default is 'roulette'; see GitHub to check other options
-                @ max_iteration_without_improv <int> - maximum number of successive iterations without improvement. If None it is ineffective
+            algorithm_parameters: AlgorithmParams object or usual dictionary with algorithm parameter;
+                it is not mandatory to provide all possible parameters
+
+        Notes:
+            - This implementation minimizes the given objective function.
+            For maximization u can multiply the function by -1 (for instance): the absolute
+                value of the output would be the actual objective function
+
+            - If u want to use set_function only and maybe u dont have usual function,
+                just set the function to something like lambda x: 0 but set function_timeout=None too
 
         for more details and examples of implementation please visit:
             https://github.com/PasaOpasen/geneticalgorithm2
@@ -137,12 +128,15 @@ class geneticalgorithm2:
 
         # all default fields
 
-        self.param: AlgorithmParams = None
         # self.crossover: Callable[[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]] = None
         # self.real_mutation: Callable[[float, float, float], float] = None
         # self.discrete_mutation: Callable[[int, int, int], int] = None
         # self.selection: Callable[[np.ndarray, int], np.ndarray] = None
 
+        self.revolution_oppositor = None
+        self.dup_oppositor = None
+        self.creator = None
+        self.init_oppositors = None
         self.f: Callable[[array1D], float] = None
         self.funtimeout: float = None
         self.set_function: Callable[[np.ndarray], np.ndarray] = None
@@ -164,9 +158,9 @@ class geneticalgorithm2:
         )
         if not isinstance(algorithm_parameters, AlgorithmParams):
             algorithm_parameters = AlgorithmParams.from_dict(algorithm_parameters)
-
-        algorithm_parameters.check_if_valid()
+        algorithm_parameters.validate()
         self.param = algorithm_parameters
+
         self.crossover, self.real_mutation, self.discrete_mutation, self.selection = algorithm_parameters.get_CMS_funcs()
 
         # dimension and area bounds
@@ -218,7 +212,8 @@ class geneticalgorithm2:
         self._set_parents_count(self.param.parents_portion)
         self._set_elit_count(self.population_size, self.param.elit_ratio)
         assert self.parents_count >= self.elit_count, (
-            f"\n number of parents ({self.parents_count}) must be greater than number of elits ({self.elit_count})"
+            f"\n number of parents ({self.parents_count}) "
+            f"must be greater than number of elits ({self.elit_count})"
         )
 
         self._set_max_iterations()
@@ -238,13 +233,12 @@ class geneticalgorithm2:
         self.indexes_int = np.array([])
         self.indexes_float = np.array([])
 
-        VALID_STRINGS = ('bool', 'int', 'real')
         assert_message = (
             f"\n variable_type must be 'bool', 'int', 'real' or a sequence with 'int' and 'real', got {variable_type}"
         )
 
         if isinstance(variable_type, str):
-            assert (variable_type in VALID_STRINGS), assert_message
+            assert (variable_type in VARIABLE_TYPE.__args__), assert_message
             if variable_type == 'real':
                 self.indexes_float = indexes
             else:
@@ -260,7 +254,7 @@ class geneticalgorithm2:
                 "don't use 'bool' if variable_type is a sequence, "
                 "for 'boolean' case use 'int' and specify boundary as (0,1)"
             )
-            assert all(v in VALID_STRINGS for v in variable_type), assert_message
+            assert all(v in VARIABLE_TYPE.__args__ for v in variable_type), assert_message
 
             vartypes = np.array(variable_type)
             self.indexes_int = indexes[vartypes == 'int']
@@ -376,9 +370,9 @@ class geneticalgorithm2:
 
         part = count / total
 
-        filled_len = round(geneticalgorithm2.PROGRESS_BAR_LEN * part)
+        filled_len = round(GeneticAlgorithm2.PROGRESS_BAR_LEN * part)
         percents = round(100.0 * part, 1)
-        bar = '|' * filled_len + '_' * (geneticalgorithm2.PROGRESS_BAR_LEN - filled_len)
+        bar = '|' * filled_len + '_' * (GeneticAlgorithm2.PROGRESS_BAR_LEN - filled_len)
 
         self.progress_stream.write('\r%s %s%s %s' % (bar, percents, '%', status))
         self.progress_stream.flush()
@@ -415,15 +409,15 @@ class geneticalgorithm2:
 
         return obj, eval_time
 
-    def _set_mutation_indexes(self, mutation_indexes: Optional[Sequence[int]]):
+    def _set_mutation_indexes(self, mutation_indexes: Optional[Iterable[int]]):
 
         if mutation_indexes is None:
             self.indexes_float_mut = self.indexes_float
             self.indexes_int_mut = self.indexes_int
         else:
             tmp_indexes = set(mutation_indexes)
-            self.indexes_int_mut = np.array(list(set(self.indexes_int).intersection(tmp_indexes)))
-            self.indexes_float_mut = np.array(list(set(self.indexes_float).intersection(tmp_indexes)))
+            self.indexes_int_mut = np.array(list(tmp_indexes.intersection(self.indexes_int)))
+            self.indexes_float_mut = np.array(list(tmp_indexes.intersection(self.indexes_float)))
 
             if self.indexes_float_mut.size == 0 and self.indexes_int_mut.size == 0:
                 warnings.warn(f"No mutation dimensions!!! Check ur mutation indexes!!")
@@ -438,28 +432,28 @@ class geneticalgorithm2:
         # deprecated
         disable_progress_bar: bool = False,
 
-        set_function: Optional[Callable[[array2D], array1D]] = None,
+        set_function: SetFunctionToMinimize = None,
         apply_function_to_parents: bool = False,
         start_generation: GenerationConvertible = Generation(),
         studEA: bool = False,
-        mutation_indexes: Optional[Union[Sequence[int], Set[int]]] = None,
+        mutation_indexes: Optional[Iterable[int]] = None,
 
-        init_creator: Optional[Callable[[], array1D]] = None,
-        init_oppositors: Optional[Sequence[Callable[[array1D], array1D]]] = None,
+        init_creator: Optional[CreatorFunc] = None,
+        init_oppositors: Optional[Sequence[OppositorFunc]] = None,
 
-        duplicates_oppositor: Optional[Callable[[array1D], array1D]] = None,
+        duplicates_oppositor: Optional[OppositorFunc] = None,
         remove_duplicates_generation_step: Optional[int] = None,
 
-        revolution_oppositor: Optional[Callable[[array1D], array1D]] = None,
+        revolution_oppositor: Optional[OppositorFunc] = None,
         revolution_after_stagnation_step: Optional[int] = None,
         revolution_part: float = 0.3,
 
         population_initializer: Tuple[
-            int, Callable[[array2D, array1D], Tuple[array2D, array1D]]
-        ] = Population_initializer(select_best_of=1, local_optimization_step='never', local_optimizer=None),
+            int, PopulationModifier
+        ] = get_population_initializer(select_best_of=1, local_optimization_step='never', local_optimizer=None),
 
         stop_when_reached: Optional[float] = None,
-        callbacks: Optional[Sequence[CallbackFunc]] = None,
+        callbacks: Optional[Sequence[SimpleCallbackFunc]] = None,
         middle_callbacks: Optional[Sequence[MiddleCallbackFunc]] = None,  #+
         time_limit_secs: Optional[float] = None,
         save_last_generation_as: Optional[str] = None,
@@ -475,17 +469,13 @@ class geneticalgorithm2:
 
             progress_bar_stream: 'stdout', 'stderr' or None to disable progress bar
 
-            disable_progress_bar:
+            disable_progress_bar: deprecated
 
-            set_function : 2D-array -> 1D-array function,
-                which applyes to matrix of population (size (samples, dimention))
-                    to estimate their values
+            set_function: set function to be used instead of usual function
 
-            apply_function_to_parents: apply function to parents from previous generation (if it's needed)
+            apply_function_to_parents: whether to apply function to parents from previous generation (if it's needed)
 
-            start_generation: Generation object or a dictionary with structure
-                {'variables':2D-array of samples, 'scores': function values on samples}
-                or path to .npz file (str) with saved generation; if 'scores' value is None the scores will be compute
+            start_generation: initial generation object of any `GenerationConvertible` type
 
             studEA: using stud EA strategy (crossover with best object always)
 
@@ -494,10 +484,12 @@ class geneticalgorithm2:
             init_creator: the function creates population samples.
                 By default -- random uniform for real variables and random uniform for int
             init_oppositors: the list of oppositors creates oppositions for base population. No by default
+
             duplicates_oppositor: oppositor for applying after duplicates removing.
                 By default -- using just random initializer from creator
             remove_duplicates_generation_step: step for removing duplicates (have a sense with discrete tasks).
                 No by default
+
             revolution_oppositor: oppositor for revolution time. No by default
             revolution_after_stagnation_step: create revolution after this generations of stagnation. No by default
             revolution_part: float, the part of generation to being oppose. By default is 0.3
@@ -510,7 +502,7 @@ class geneticalgorithm2:
             callbacks: sequence of callback functions with structure:
                 (generation_number, report_list, last_population, last_scores) -> do some action
 
-            middle_callbacks: sequence of functions made MiddleCallbacks class
+            middle_callbacks: sequence of functions made `MiddleCallback` class
 
             time_limit_secs: limit time of working (in seconds)
 
@@ -578,11 +570,7 @@ class geneticalgorithm2:
         pop: array2D = None
         scores: array1D = None
 
-        #
-        #
-        # callbacks part
-        #
-        #
+        #region CALLBACKS
 
         def get_data():
             """
@@ -667,7 +655,7 @@ class geneticalgorithm2:
                 if flag:
                     set_data(data)  # update global date if there was real callback step
 
-        ############################################################
+        #endregion
 
         start_time = time.time()
         time_is_done = (
@@ -676,9 +664,9 @@ class geneticalgorithm2:
             else (lambda: int(time.time() - start_time) >= time_limit_secs)
         )
 
-        ############################################################# 
-        # Initial Population
-        self.set_function = set_function or geneticalgorithm2.default_set_function(self.f)
+        self.set_function = set_function or GeneticAlgorithm2.default_set_function(self.f)
+
+        #region Initial population, duplicates filter, revolutionary
 
         pop_coef, initializer_func = population_initializer
         
@@ -826,6 +814,8 @@ class geneticalgorithm2:
                 args = args[:scores.size]
                 return combined[args], combined_scores[args]
 
+        #enregion
+
         #
         #
         #  START ALGORITHM LOGIC
@@ -1047,12 +1037,14 @@ class geneticalgorithm2:
             last_generation.save(save_last_generation_as)
 
         if enable_printing:
-            show = ' '*200
-            sys.stdout.write(f'\r{show}\n')
-            sys.stdout.write(f'\r The best found solution:\n {pop[0]}')
-            sys.stdout.write(f'\n\n Objective function:\n {self.best_function}\n')
-            sys.stdout.write(f'\n Used generations: {len(self.report)}')
-            sys.stdout.write(f'\n Used time: {time.time() - start_time:.3g} seconds\n')
+            show = ' ' * 200
+            sys.stdout.write(
+                f'\r{show}\n'
+                f'\r The best found solution:\n {pop[0]}'
+                f'\n\n Objective function:\n {self.best_function}\n'
+                f'\n Used generations: {len(self.report)}'
+                f'\n Used time: {time.time() - start_time:.3g} seconds\n'
+            )
             sys.stdout.flush() 
         
         if not no_plot:
